from .optimizer import *  # noqa
from .collapser import *  # noqa
from .molecule import *  # noqa
from .atom import *  # noqa
from .bond import *  # noqa
from .utilities import *  # noqa
from .results import *  # noqa



import pytest
import numpy as np
import mchammer as mch


@pytest.fixture(
    params=(
        (mch.Atom(id=0, element_string='N'), 0, 'N'),
        (mch.Atom(id=65, element_string='P'), 65, 'P'),
        (mch.Atom(id=2, element_string='C'), 2, 'C'),
    )
)
def atom_info(request):
    return request.param


@pytest.fixture(
    params=(
        (mch.Bond(id=0, atom_ids=(0, 1)), 0, 0, 1),
        (mch.Bond(id=65, atom_ids=(2, 3)), 65, 2, 3),
        (mch.Bond(id=2, atom_ids=(3, 4)), 2, 3, 4),
        (mch.Bond(id=3, atom_ids=(0, 9)), 3, 0, 9),
    )
)
def bond_info(request):
    return request.param


@pytest.fixture
def atoms():
    return [
        mch.Atom(0, 'C'), mch.Atom(1, 'C'), mch.Atom(2, 'C'),
        mch.Atom(3, 'C'), mch.Atom(4, 'C'), mch.Atom(5, 'C'),
    ]


@pytest.fixture
def bonds():
    return [
        mch.Bond(0, (0, 1)), mch.Bond(1, (0, 2)), mch.Bond(2, (0, 3)),
        mch.Bond(3, (3, 4)), mch.Bond(4, (3, 5))
    ]


@pytest.fixture
def position_matrix():
    return np.array([
        [0, 1, 0],
        [1, 1, 0],
        [-1, 1, 0],
        [0, 10, 0],
        [1, 10, 0],
        [-1, 10, 0],
    ])


@pytest.fixture
def position_matrix2():
    return np.array([
        [0, 1, 0],
        [1, 1, 0],
        [-1, 1, 0],
        [0, 20, 0],
        [1, 20, 0],
        [-1, 20, 0],
    ])


@pytest.fixture
def centroid():
    return np.array([0, 5.5, 0])


@pytest.fixture
def molecule(atoms, bonds, position_matrix):
    return mch.Molecule(
        atoms=atoms,
        bonds=bonds,
        position_matrix=position_matrix
    )


@pytest.fixture
def bond_vector():
    return np.array([0, 9, 0])


@pytest.fixture
def bond_potentials():
    return [50, 0, 50, 200, 450, 800, 1250]


@pytest.fixture
def nonbond_potentials():
    return [
        34.559999999999995, 4.319999999999999, 1.2799999999999998,
        0.5399999999999999, 0.27647999999999995, 0.15999999999999998,
        0.10075801749271138,
    ]


@pytest.fixture
def nonbonded_potential():
    return 147.2965949864993


@pytest.fixture
def system_potential():
    return 2597.2965949864993


@pytest.fixture
def subunits():
    return {0: {0, 1, 2}, 1: {3, 4, 5}}


@pytest.fixture
def position_matrix3():
    return np.array([
        [0, 1, 0],
        [1, 1, 0],
        [-1, 1, 0],
        [0, 5, 0],
        [1, 5, 0],
        [-1, 5, 0],
    ])


@pytest.fixture
def optimizer():
    return mch.Optimizer(
        step_size=0.1,
        target_bond_length=2.0,
        num_steps=100
    )


@pytest.fixture
def coll_atoms():
    return [
        mch.Atom(0, 'C'), mch.Atom(1, 'C'), mch.Atom(2, 'C'),
        mch.Atom(3, 'C'), mch.Atom(4, 'C'), mch.Atom(5, 'C'),
    ]


@pytest.fixture
def coll_bonds():
    return [
        mch.Bond(0, (0, 1)), mch.Bond(1, (0, 2)), mch.Bond(2, (0, 3)),
        mch.Bond(3, (3, 4)), mch.Bond(4, (3, 5))
    ]


@pytest.fixture
def coll_position_matrix():
    return np.array([
        [0, 1, 0],
        [1, 1, 0],
        [-1, 1, 0],
        [0, 10, 0],
        [1, 10, 0],
        [-1, 10, 0],
    ])


@pytest.fixture
def coll_vectors(request):
    return {0: np.array([2, -1.5, 0]), 1: np.array([0, 2.5, 0])}


@pytest.fixture
def coll_su_dists(request):
    return [
        9, 9.055385138137417, 9.055385138137417, 9.055385138137417,
        9, 9.219544457292887, 9.055385138137417, 9.219544457292887,
        9,
    ]


@pytest.fixture
def coll_scales(request):
    return {0: 0.2, 1: 1.0}


@pytest.fixture
def coll_step(request):
    return 1.5


@pytest.fixture
def coll_position_matrix2():
    return np.array([
        [-0.6, 1.45, 0],
        [0.4, 1.45, 0],
        [-1.6, 1.45, 0],
        [0, 6.25, 0],
        [1, 6.25, 0],
        [-1, 6.25, 0],
    ])


@pytest.fixture
def su_vectors(request):
    return {0: np.array([0, -4.5, 0]), 1: np.array([0, 4.5, 0])}


@pytest.fixture
def su_scales(request):
    return {0: 1.0, 1: 1.0}


@pytest.fixture
def coll_molecule(coll_atoms, coll_bonds, coll_position_matrix):
    return mch.Molecule(
        atoms=coll_atoms,
        bonds=coll_bonds,
        position_matrix=coll_position_matrix
    )


@pytest.fixture
def coll_subunits():
    return {0: {0, 1, 2}, 1: {3, 4, 5}}


@pytest.fixture
def coll_final_position_matrix():
    return np.array([
        [0., 4.75262477, 0.],
        [1., 4.75262477, 0.],
        [-1., 4.75262477, 0.],
        [0., 6.24737523, 0.],
        [1., 6.24737523, 0.],
        [-1., 6.24737523, 0.],
    ])


@pytest.fixture
def collapser():
    return mch.Collapser(
        step_size=0.05,
        distance_threshold=1.5,
        scale_steps=True,
    )



"""
Results
=======

#. :class:`.StepResult`
#. :class:`.MCStepResult`
#. :class:`.Result`

Classes for maintaining optimization results.

"""


class StepResult:
    """
    Results of a step.

    """

    def __init__(self, step, position_matrix, max_bond_distance, log):
        """
        Initialize a :class:`StepResult` instance.

        Parameters
        ----------
        step : :class:`int`
            Step number.

        position_matrix : :class:`numpy.ndarray`
            A position matrix after performing this step. The shape of
            the matrix is ``(n, 3)``.

        max_bond_distance : :class:`float`
            Max length of bonds to be optimized in Angstrom.

        log : :class:`str`
            String log of this step.

        """

        self._step = step
        self._position_matrix = position_matrix
        self._max_bond_distance = max_bond_distance
        self._log = log

    def get_step(self):
        return self._step

    def get_log(self):
        return self._log

    def get_position_matrix(self):
        return self._position_matrix

    def get_max_bond_distance(self):
        return self._max_bond_distance


class MCStepResult(StepResult):
    """
    Results of a step.

    """

    def __init__(
        self,
        step,
        position_matrix,
        passed,
        system_potential,
        nonbonded_potential,
        max_bond_distance,
        log,
    ):
        """
        Initialize a :class:`MCStepResult` instance.

        Parameters
        ----------
        step : :class:`int`
            Step number.

        position_matrix : :class:`numpy.ndarray`
            A position matrix after performing this step. The shape of
            the matrix is ``(n, 3)``.

        passed : :class:`bool` or :class:`None`
            Flag for whether the MC move passed, or was reverted to
            the previous step.

        system_potential : :class:`float`
            System potential of the structure after this step.

        nonbonded_potential : :class:`float`
            Nonbonded potential of the structure after this step.

        max_bond_distance : :class:`float`
            Max length of bonds to be optimized in Angstrom.

        log : :class:`str`
            String log of this step.

        """

        self._step = step
        self._log = ''
        self._position_matrix = position_matrix
        self._passed = passed
        self._system_potential = system_potential
        self._nonbonded_potential = nonbonded_potential
        self._max_bond_distance = max_bond_distance

    def get_passed(self):
        return self._passed

    def get_system_potential(self):
        return self._system_potential

    def get_nonbonded_potential(self):
        return self._nonbonded_potential

    def get_properties(self):
        return {
            'max_bond_distance': self._max_bond_distance,
            'system_potential': self._system_potential,
            'nonbonded_potential': self._nonbonded_potential,
            'passed': self._passed,
        }


class Result:
    """
    Result of optimization.

    """

    def __init__(self, start_time):
        """
        Initialize a :class:`Result` instance.

        Parameters
        ----------
        start_time : :class:`time.time()`
            Start of run timing.

        """

        self._start_time = start_time
        self._log = ''
        self._step_results = {}
        self._step_count = None

    def add_step_result(self, step_result):
        """
        Add StepResult.

        """

        self._step_results[step_result.get_step()] = step_result
        self.update_log(step_result.get_log())
        self._step_count = step_result.get_step()

    def update_log(self, string):
        """
        Update result log.

        """
        self._log += string

    def get_log(self):
        """
        Get result log.

        """
        return self._log

    def get_number_passed(self):
        """
        Get Number of steps that passed.

        """
        return len([
            1 for step in self._step_results
            if self._step_results[step].get_passed()
        ])

    def get_final_position_matrix(self):
        """
        Get final molecule in result.

        """

        return (
            self._step_results[self._step_count].get_position_matrix()
        )

    def get_timing(self, time):
        """
        Get run timing.

        """
        return time - self._start_time

    def get_step_count(self):
        """
        Get step count.

        """
        return self._step_count

    def get_steps_properties(self):
        """
        Yield properties of all steps.

        """

        for step in self._step_results:
            yield step, self._step_results[step].get_properties()

    def get_trajectory(self):
        """
        Yield .Molecule of all steps.

        """

        for step in self._step_results:
            yield step, self._step_results[step].get_position_matrix()



import numpy as np
from mchammer import get_atom_distance


def test_opt_get_bond_vector(
    optimizer, bond_vector, position_matrix
):
    assert np.all(np.equal(
        bond_vector,
        optimizer._get_bond_vector(position_matrix, (0, 3)),
    ))


def test_opt_bond_potential(optimizer, bond_potentials):
    for i, d in enumerate([1, 2, 3, 4, 5, 6, 7]):
        test = optimizer._bond_potential(distance=d)
        assert np.isclose(test, bond_potentials[i], atol=1E-5)


def test_opt_nonbond_potential(optimizer, nonbond_potentials):
    for i, d in enumerate([1, 2, 3, 4, 5, 6, 7]):
        test = optimizer._nonbond_potential(distance=d)
        assert np.isclose(test, nonbond_potentials[i], atol=1E-5)


def test_opt_compute_nonbonded_potential(
    optimizer, position_matrix, nonbonded_potential
):
    test = optimizer._compute_nonbonded_potential(position_matrix)
    assert test == nonbonded_potential


def test_opt_compute_potential(
    optimizer, molecule, nonbonded_potential, system_potential
):
    test_system_potential, test_nonbond_potential = (
        optimizer._compute_potential(
            molecule, bond_pair_ids=((0, 3), ),
        )
    )
    assert test_nonbond_potential == nonbonded_potential
    assert test_system_potential == system_potential


def test_opt_translate_atoms_along_vector(
    optimizer, molecule, position_matrix, position_matrix3
):
    molecule = molecule.with_position_matrix(position_matrix)
    new_molecule = optimizer._translate_atoms_along_vector(
        mol=molecule,
        atom_ids=(3, 4, 5),
        vector=np.array([0, 5, 0])
    )
    assert np.all(np.equal(
        position_matrix3,
        new_molecule.get_position_matrix(),
    ))
    new_molecule = optimizer._translate_atoms_along_vector(
        mol=new_molecule,
        atom_ids=(3, 4, 5),
        vector=np.array([0, -5, 0])
    )
    print(position_matrix, new_molecule.get_position_matrix())
    assert np.all(np.equal(
        position_matrix,
        new_molecule.get_position_matrix(),
    ))


def test_opt_test_move(optimizer):
    # Do not test random component.
    assert optimizer._test_move(curr_pot=-1, new_pot=-2)


def test_opt_get_result(optimizer, molecule):
    original_pos_mat = molecule.get_position_matrix()
    subunits = molecule.get_subunits(bond_pair_ids=((0, 3), ))
    mol, results = optimizer.get_result(
        mol=molecule,
        bond_pair_ids=((0, 3), ),
        subunits=subunits,
    )

    final_bond_length = np.linalg.norm(
        optimizer._get_bond_vector(
            position_matrix=results.get_position_matrix(),
            bond_pair=(0, 3),
        ),
    )
    # Give it some wiggle room.
    assert 1.5 < final_bond_length
    assert final_bond_length < 2.5

    # Test all other bond lengths are equivalent.
    for bond in molecule.get_bonds():
        if (bond.get_atom1_id(), bond.get_atom2_id()) != (0, 3):
            test = get_atom_distance(
                position_matrix=results.get_position_matrix(),
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            bond_length = get_atom_distance(
                position_matrix=original_pos_mat,
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            assert np.isclose(test, bond_length, atol=1E-6)


def test_opt_get_trajectory(optimizer, molecule):
    original_pos_mat = molecule.get_position_matrix()
    subunits = molecule.get_subunits(bond_pair_ids=((0, 3), ))
    mol, results = optimizer.get_trajectory(
        mol=molecule,
        bond_pair_ids=((0, 3), ),
        subunits=subunits,
    )

    assert results.get_step_count() == 99
    assert len(tuple(results.get_steps_properties())) == 100

    final_bond_length = np.linalg.norm(
        optimizer._get_bond_vector(
            position_matrix=results.get_final_position_matrix(),
            bond_pair=(0, 3),
        ),
    )
    # Give it some wiggle room.
    assert 1.5 < final_bond_length
    assert final_bond_length < 2.5

    # Test all other bond lengths are equivalent.
    for bond in molecule.get_bonds():
        if (bond.get_atom1_id(), bond.get_atom2_id()) != (0, 3):
            test = get_atom_distance(
                position_matrix=results.get_final_position_matrix(),
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            bond_length = get_atom_distance(
                position_matrix=original_pos_mat,
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            assert np.isclose(test, bond_length, atol=1E-6)



"""
MCHammer Optimizer
==================

#. :class:`.Optimizer`

Optimizer for minimising intermolecular distances.

"""

import numpy as np
import time

from scipy.spatial.distance import pdist
from copy import deepcopy
import random

from .results import Result, MCStepResult
from .utilities import get_atom_distance


class Optimizer:
    """
    Optimize target bonds using MC algorithm.

    A Metropolis MC algorithm is applied to perform rigid
    translations of the subunits separatred by the target bonds.

    """

    def __init__(
        self,
        step_size,
        target_bond_length,
        num_steps,
        bond_epsilon=50,
        nonbond_epsilon=20,
        nonbond_sigma=1.2,
        nonbond_mu=3,
        beta=2,
        random_seed=1000,
    ):
        """
        Initialize a :class:`Optimizer` instance.

        Parameters
        ----------
        step_size : :class:`float`
            The relative size of the step to take during step.

        target_bond_length : :class:`float`
            Target equilibrium bond length for long bonds to minimize
            to.

        num_steps : :class:`int`
            Number of MC moves to perform.

        bond_epsilon : :class:`float`, optional
            Value of epsilon used in the bond potential in MC moves.
            Determines strength of the bond potential.
            Defaults to 50.

        nonbond_epsilon : :class:`float`, optional
            Value of epsilon used in the nonbond potential in MC moves.
            Determines strength of the nonbond potential.
            Defaults to 20.

        nonbond_sigma : :class:`float`, optional
            Value of sigma used in the nonbond potential in MC moves.
            Defaults to 1.2.

        nonbond_mu : :class:`float`, optional
            Value of mu used in the nonbond potential in MC moves.
            Determines the steepness of the nonbond potential.
            Defaults to 3.

        beta : :class:`float`, optional
            Value of beta used in the in MC moves. Beta takes the
            place of the inverse boltzmann temperature.
            Defaults to 2.

        random_seed : :class:`int` or :class:`NoneType`, optional
            Random seed to use for MC algorithm. Should only be set to
            ``None`` if system-based random seed is desired. Defaults
            to 1000.

        """

        self._step_size = step_size
        self._target_bond_length = target_bond_length
        self._num_steps = num_steps
        self._bond_epsilon = bond_epsilon
        self._nonbond_epsilon = nonbond_epsilon
        self._nonbond_sigma = nonbond_sigma
        self._nonbond_mu = nonbond_mu
        self._beta = beta
        if random_seed is None:
            random.seed()
            np.random.seed()
        else:
            random.seed(random_seed)
            np.random.seed(random_seed)

    def _get_bond_vector(self, position_matrix, bond_pair):
        """
        Get vector from atom1 to atom2 in bond.

        """

        atom1_pos = position_matrix[bond_pair[0]]
        atom2_pos = position_matrix[bond_pair[1]]
        return atom2_pos - atom1_pos

    def _bond_potential(self, distance):
        """
        Define an arbitrary parabolic bond potential.

        This potential has no relation to an empircal forcefield.

        """

        potential = (distance - self._target_bond_length) ** 2
        potential = self._bond_epsilon * potential

        return potential

    def _nonbond_potential(self, distance):
        """
        Define an arbitrary repulsive nonbonded potential.

        This potential has no relation to an empircal forcefield.

        """

        return (
            self._nonbond_epsilon * (
                (self._nonbond_sigma/distance) ** self._nonbond_mu
            )
        )

    def _compute_nonbonded_potential(self, position_matrix):
        # Get all pairwise distances between atoms in each subunut.
        pair_dists = pdist(position_matrix)
        nonbonded_potential = np.sum(
            self._nonbond_potential(pair_dists)
        )

        return nonbonded_potential

    def _compute_potential(self, mol, bond_pair_ids):
        position_matrix = mol.get_position_matrix()
        nonbonded_potential = self._compute_nonbonded_potential(
            position_matrix=position_matrix,
        )
        system_potential = nonbonded_potential
        for bond in bond_pair_ids:
            system_potential += self._bond_potential(
                distance=get_atom_distance(
                    position_matrix=position_matrix,
                    atom1_id=bond[0],
                    atom2_id=bond[1],
                )
            )

        return system_potential, nonbonded_potential

    def _translate_atoms_along_vector(self, mol, atom_ids, vector):

        new_position_matrix = deepcopy(mol.get_position_matrix())
        for atom in mol.get_atoms():
            if atom.get_id() not in atom_ids:
                continue
            pos = mol.get_position_matrix()[atom.get_id()]
            new_position_matrix[atom.get_id()] = pos - vector

        mol = mol.with_position_matrix(new_position_matrix)
        return mol

    def _test_move(self, curr_pot, new_pot):

        if new_pot < curr_pot:
            return True
        else:
            exp_term = np.exp(-self._beta*(new_pot-curr_pot))
            rand_number = random.random()

            if exp_term > rand_number:
                return True
            else:
                return False

    def _output_top_lines(self):

        string = (
            '====================================================\n'
            '                MCHammer optimisation               \n'
            '                ---------------------               \n'
            '                    Andrew Tarzia                   \n'
            '                ---------------------               \n'
            '                                                    \n'
            'Settings:                                           \n'
            f' step size = {self._step_size} \n'
            f' target bond length = {self._target_bond_length} \n'
            f' num. steps = {self._num_steps} \n'
            f' bond epsilon = {self._bond_epsilon} \n'
            f' nonbond epsilon = {self._nonbond_epsilon} \n'
            f' nonbond sigma = {self._nonbond_sigma} \n'
            f' nonbond mu = {self._nonbond_mu} \n'
            f' beta = {self._beta} \n'
            '====================================================\n\n'
        )

        return string

    def _run_first_step(
        self,
        mol,
        bond_pair_ids,
        subunits,
    ):

        system_potential, nonbonded_potential = (
            self._compute_potential(
                mol=mol,
                bond_pair_ids=bond_pair_ids
            )
        )

        # Update properties at each step.
        max_bond_distance = max([
            get_atom_distance(
                position_matrix=mol.get_position_matrix(),
                atom1_id=bond[0],
                atom2_id=bond[1],
            )
            for bond in bond_pair_ids
        ])
        step_result = MCStepResult(
            step=0,
            position_matrix=mol.get_position_matrix(),
            passed=None,
            system_potential=system_potential,
            nonbonded_potential=nonbonded_potential,
            max_bond_distance=max_bond_distance,
            log=(
                f"{0} "
                f"{system_potential} "
                f"{nonbonded_potential} "
                f"{max_bond_distance} "
                '-- --\n'
            ),
        )

        return mol, step_result

    def _run_step(
        self,
        mol,
        bond_pair_ids,
        subunits,
        step,
        system_potential,
        nonbonded_potential,
    ):

        position_matrix = mol.get_position_matrix()

        # Randomly select a bond to optimize from bonds.
        bond_ids = random.choice(bond_pair_ids)
        bond_vector = self._get_bond_vector(
            position_matrix=position_matrix,
            bond_pair=bond_ids
        )

        # Get subunits connected by selected bonds.
        subunit_1 = [
            i for i in subunits if bond_ids[0] in subunits[i]
        ][0]
        subunit_2 = [
            i for i in subunits if bond_ids[1] in subunits[i]
        ][0]

        # Choose subunit to move out of the two connected by the
        # bond randomly.
        moving_su = random.choice([subunit_1, subunit_2])
        moving_su_atom_ids = tuple(i for i in subunits[moving_su])

        # Random number from -1 to 1 for multiplying translation.
        rand = (random.random() - 0.5) * 2
        # Define translation along long bond vector where
        # direction is from force, magnitude is randomly
        # scaled.
        bond_translation = -bond_vector * self._step_size * rand

        # Define subunit COM vector to molecule COM.
        cent = mol.get_centroid()
        su_cent_vector = (
            mol.get_centroid(atom_ids=moving_su_atom_ids)-cent
        )
        com_translation = su_cent_vector * self._step_size * rand

        # Randomly choose between translation along long bond
        # vector or along BB-COM vector.
        translation_vector = random.choice([
            bond_translation,
            com_translation,
        ])

        # Translate building block.
        # Update atom position of building block.
        mol = self._translate_atoms_along_vector(
            mol=mol,
            atom_ids=moving_su_atom_ids,
            vector=translation_vector,
        )

        new_system_potential, new_nonbonded_potential = (
            self._compute_potential(
                mol=mol,
                bond_pair_ids=bond_pair_ids
            )
        )

        if self._test_move(system_potential, new_system_potential):
            updated = 'T'
            passed = True
            system_potential = new_system_potential
            nonbonded_potential = new_nonbonded_potential
        else:
            updated = 'F'
            passed = False
            # Reverse move.
            mol = self._translate_atoms_along_vector(
                mol=mol,
                atom_ids=moving_su_atom_ids,
                vector=-translation_vector,
            )

        # Update properties at each step.
        max_bond_distance = max([
            get_atom_distance(
                position_matrix=mol.get_position_matrix(),
                atom1_id=bond[0],
                atom2_id=bond[1],
            )
            for bond in bond_pair_ids
        ])
        step_result = MCStepResult(
            step=step,
            position_matrix=mol.get_position_matrix(),
            passed=passed,
            system_potential=system_potential,
            nonbonded_potential=nonbonded_potential,
            max_bond_distance=max_bond_distance,
            log=(
                f"{step} "
                f"{system_potential} "
                f"{nonbonded_potential} "
                f"{max_bond_distance} "
                f'{bond_ids} {updated}\n'
            ),
        )

        return mol, step_result

    def get_trajectory(self, mol, bond_pair_ids, subunits):
        """
        Get trajectory of optimization run on `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        bond_pair_ids :
            :class:`iterable` of :class:`tuple` of :class:`ints`
            Iterable of pairs of atom ids with bond between them to
            optimize.

        subunits : :class:`.dict`
            The subunits of `mol` split by bonds defined by
            `bond_pair_ids`. Key is subunit identifier, Value is
            :class:`iterable` of atom ids in subunit.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        result : :class:`.Result`
            The result of the optimization including all steps.

        """

        result = Result(start_time=time.time())

        result.update_log(self._output_top_lines())
        result.update_log(
            f'There are {len(bond_pair_ids)} bonds to optimize.\n'
        )
        result.update_log(
            f'There are {len(subunits)} sub units with N atoms:\n'
            f'{[len(subunits[i]) for i in subunits]}\n'
        )
        result.update_log(
            '====================================================\n'
            '                 Running optimisation!              \n'
            '====================================================\n\n'
        )
        result.update_log(
            'step system_potential nonbond_potential max_dist '
            'opt_bbs updated?\n'
        )
        mol, step_result = self._run_first_step(
            mol,
            bond_pair_ids,
            subunits,
        )
        system_potential = step_result.get_system_potential()
        nonbonded_potential = step_result.get_nonbonded_potential()
        result.add_step_result(step_result=step_result)

        for step in range(1, self._num_steps):
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
                system_potential=system_potential,
                nonbonded_potential=nonbonded_potential,
            )

            system_potential = step_result.get_system_potential()
            nonbonded_potential = step_result.get_nonbonded_potential()
            result.add_step_result(step_result=step_result)

        num_passed = result.get_number_passed()
        result.update_log(
            string=(
                '\n============================================\n'
                'Optimisation done:\n'
                f'{num_passed} steps passed: '
                f'{(num_passed/self._num_steps)*100}'
                '%\n'
                f'Total optimisation time: '
                f'{round(result.get_timing(time.time()), 4)}s\n'
                '============================================\n'
            ),
        )

        return mol, result

    def get_result(self, mol, bond_pair_ids, subunits):
        """
        Get final result of optimization run on `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        bond_pair_ids :
            :class:`iterable` of :class:`tuple` of :class:`ints`
            Iterable of pairs of atom ids with bond between them to
            optimize.

        subunits : :class:`.dict`
            The subunits of `mol` split by bonds defined by
            `bond_pair_ids`. Key is subunit identifier, Value is
            :class:`iterable` of atom ids in subunit.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        result : :class:`.MCStepResult`
            The result of the final optimization step.

        """

        mol, step_result = self._run_first_step(
            mol,
            bond_pair_ids,
            subunits,
        )
        system_potential = step_result.get_system_potential()
        nonbonded_potential = step_result.get_nonbonded_potential()

        for step in range(1, self._num_steps):
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
                system_potential=system_potential,
                nonbonded_potential=nonbonded_potential,
            )
            system_potential = step_result.get_system_potential()
            nonbonded_potential = step_result.get_nonbonded_potential()

        return mol, step_result



import pytest
import os
import numpy as np


def test_molecule_get_position_matrix(molecule, position_matrix):
    assert np.all(np.equal(
        position_matrix,
        molecule.get_position_matrix(),
    ))


def test_molecule_with_position_matrix(
    molecule, position_matrix2, position_matrix
):
    test = molecule.with_position_matrix(position_matrix2)
    assert np.all(np.equal(
        position_matrix2,
        test.get_position_matrix(),
    ))


@pytest.fixture
def path(request, tmpdir):
    return os.path.join(tmpdir, 'molecule.xyz')


def test_molecule_write_xyz_file(molecule, path):
    molecule.write_xyz_file(path)
    content = molecule._write_xyz_content()
    with open(path, 'r') as f:
        test_lines = f.readlines()

    assert ''.join(test_lines) == ''.join(content)


def test_molecule_get_atoms(molecule, atoms):
    for test, atom in zip(molecule.get_atoms(), atoms):
        assert test.get_id() == atom.get_id()
        assert test.get_element_string() == atom.get_element_string()


def test_molecule_get_bonds(molecule, bonds):
    for test, bond in zip(molecule.get_bonds(), bonds):
        assert test.get_id() == bond.get_id()
        assert test.get_atom1_id() == bond.get_atom1_id()
        assert test.get_atom2_id() == bond.get_atom2_id()


def test_molecule_get_centroid(
    molecule, position_matrix, centroid
):
    test = molecule.with_position_matrix(position_matrix)
    assert np.all(np.allclose(
        centroid,
        test.get_centroid(),
        atol=1E-6,
    ))


def test_molecule_get_subunits(molecule, subunits):
    test = molecule.get_subunits(bond_pair_ids=((0, 3), ))
    assert test == subunits



"""
MCHammer Collapser
==================

#. :class:`.Collapser`

Optimizer for minimising intermolecular distances.

"""

import numpy as np
import time

from itertools import combinations, product

from .results import Result, StepResult
from .utilities import get_atom_distance


class Collapser:
    """
    Moves rigid-subunits toward center of mass of molecule.

    """

    def __init__(
        self,
        step_size,
        distance_threshold,
        scale_steps=True,
    ):
        """
        Initialize a :class:`Collapser` instance.

        Parameters
        ----------
        step_size : :class:`float`
            The relative size of the step to take during collapse.

        distance_threshold : :class:`float`
            Distance between distinct subunits to use as
            threshold for halting collapse in Angstrom.

        scale_steps : :class:`bool`, optional
            Whether to scale the step of each distict building block
            by their relative distance from the molecules centroid.
            Defaults to ``True``

        """

        self._step_size = step_size
        self._distance_threshold = distance_threshold
        self._scale_steps = scale_steps

    def _has_short_contacts(self, mol, subunits):
        """
        Calculate if there are short contants in mol.

        """

        return any(
            dist < self._distance_threshold
            for dist in self._get_subunit_distances(mol, subunits)
        )

    def _get_subunit_distances(self, mol, subunits):
        """
        Yield the distances between subunits in mol.

        Ignores H atoms.

        """

        atom_elements = [
            i.get_element_string() for i in mol.get_atoms()
        ]
        position_matrix = mol.get_position_matrix()

        for su1, su2 in combinations(subunits, 2):
            if su1 == su2:
                continue
            su1_atom_ids = subunits[su1]
            su2_atom_ids = subunits[su2]
            for atom1_id, atom2_id in product(
                su1_atom_ids, su2_atom_ids
            ):
                atom1_element = atom_elements[atom1_id]
                atom2_element = atom_elements[atom2_id]
                if atom1_element != 'H' and atom2_element != 'H':
                    yield get_atom_distance(
                        position_matrix=position_matrix,
                        atom1_id=atom1_id,
                        atom2_id=atom2_id,
                    )

    def _get_new_position_matrix(
        self,
        mol,
        subunits,
        vectors,
        scales,
        step_size,
    ):
        """
        Get the position matrix of the mol after translation.

        """

        if step_size is None:
            step_size = self._step_size

        new_position_matrix = mol.get_position_matrix()
        for su in subunits:
            for atom_id in subunits[su]:
                pos = mol.get_position_matrix()[atom_id]
                new_position_matrix[atom_id] = (
                    pos - step_size*vectors[su]*scales[su]
                )

        return new_position_matrix

    def _get_subunit_vectors(self, mol, subunits):
        """
        Get the subunit to COM vectors.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        subunits : :class:`.dict`
            The subunits of `mol` split by bonds defined by
            `bond_pair_ids`. Key is subunit identifier, Value is
            :class:`iterable` of atom ids in subunit.

        Returns
        -------
        su_cent_vectors :
            :class:`dict` mapping :class:`int`: to
            :class:`numpy.ndarray`
            Dictionary mapping subunit ids (keys) to centroid
            vectors (values) of each distinct building block in the
            molecule.

        su_cent_scales :
            :class:`dict` mapping :class:`int`: to :class:`float`
            Dictionary mapping subunit ids (keys) to relative
            magnitude of centroid vectors (values) of each distinct
            building block in the molecule.

        """

        centroid = mol.get_centroid()

        # Get subunit centroid to molecule centroid vector.
        su_cent_vectors = {
            i: mol.get_centroid(atom_ids=subunits[i])-centroid
            for i in subunits
        }

        # Scale the step size based on the different distances of
        # subunits from the COM. Impacts anisotropic topologies.
        if self._scale_steps:
            norms = {
                i: np.linalg.norm(su_cent_vectors[i])
                for i in su_cent_vectors
            }
            max_distance = max(list(norms.values()))
            su_cent_scales = {
                i: norms[i]/max_distance
                for i in norms
            }
        else:
            su_cent_scales = {
                i: 1
                for i in su_cent_vectors
            }

        return su_cent_vectors, su_cent_scales

    def _run_step(
        self,
        mol,
        bond_pair_ids,
        subunits,
        step,
        step_size=None,
    ):

        # Get the subunit to centroid vectors and their relative scale.
        su_cent_vectors, su_cent_scales = self._get_subunit_vectors(
            mol=mol,
            subunits=subunits
        )

        new_position_matrix = self._get_new_position_matrix(
            mol=mol,
            subunits=subunits,
            vectors=su_cent_vectors,
            scales=su_cent_scales,
            step_size=step_size,
        )

        mol = mol.with_position_matrix(new_position_matrix)
        # Update properties at each step.
        max_bond_distance = max([
            get_atom_distance(
                position_matrix=mol.get_position_matrix(),
                atom1_id=bond[0],
                atom2_id=bond[1],
            )
            for bond in bond_pair_ids
        ])
        step_result = StepResult(
            step=step,
            position_matrix=new_position_matrix,
            max_bond_distance=max_bond_distance,
            log=f"{step} {max_bond_distance}\n"
        )
        return mol, step_result

    def get_trajectory(self, mol, bond_pair_ids, subunits):
        """
        Get trajectory of optimization run on `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        bond_pair_ids :
            :class:`iterable` of :class:`tuple` of :class:`ints`
            Iterable of pairs of atom ids with bond between them to
            optimize.

        subunits : :class:`.dict`
            The subunits of `mol` split by bonds defined by
            `bond_pair_ids`. Key is subunit identifier, Value is
            :class:`iterable` of atom ids in subunit.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        result : :class:`.Result`
            The result of the optimization including all steps.

        """

        result = Result(start_time=time.time())
        result.update_log(
            '====================================================\n'
            f"                Collapser algorithm.\n"
            '====================================================\n'
            f"Step size: {self._step_size}\n"
            f"Scale steps?: {self._scale_steps}\n"
            f"Distance threshold: {self._distance_threshold}\n"
        )
        result.update_log(
            f'There are {len(bond_pair_ids)} bonds to optimize.\n'
        )
        result.update_log(
            f'There are {len(subunits)} sub units with N atoms:\n'
            f'{[len(subunits[i]) for i in subunits]}\n'
        )
        result.update_log(
            '====================================================\n'
            '                 Running optimisation!              \n'
            '====================================================\n\n'
        )

        step = 0
        while not self._has_short_contacts(mol, subunits):
            step += 1
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
            )
            result.add_step_result(step_result=step_result)

        # Check that we have not gone too far.
        min_dist = min(
            dist for dist in self._get_subunit_distances(mol, subunits)
        )
        if min_dist < self._distance_threshold / 2:
            step += 1
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
                step_size=-(self._step_size/2),
            )
            result.add_step_result(step_result=step_result)

        result.update_log(
            '====================================================\n'
            f"Steps run: {step}\n"
            f"Minimum inter-subunit distance: {min_dist}\n"
            '====================================================\n'
        )

        return mol, result

    def get_result(self, mol, bond_pair_ids, subunits):
        """
        Get final result of the optimization of `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        bond_pair_ids :
            :class:`iterable` of :class:`tuple` of :class:`ints`
            Iterable of pairs of atom ids with bond between them to
            optimize.

        subunits : :class:`.dict`
            The subunits of `mol` split by bonds defined by
            `bond_pair_ids`. Key is subunit identifier, Value is
            :class:`iterable` of atom ids in subunit.

        Returns
        -------
        mol : :class:`.Molecule`
            The optimized molecule.

        result : :class:`.StepResult`
            The result of the final optimization step.

        """

        step = 0
        while not self._has_short_contacts(mol, subunits):
            step += 1
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
            )

        # Check that we have not gone too far.
        min_dist = min(
            dist for dist in self._get_subunit_distances(mol, subunits)
        )
        if min_dist < self._distance_threshold / 2:
            step += 1
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
                step_size=-(self._step_size/2),
            )
        return mol, step_result



"""
This module defines general-purpose objects, functions and classes.

"""

from scipy.spatial.distance import euclidean


def get_atom_distance(position_matrix, atom1_id, atom2_id):
    """
    Return the distance between two atoms.

    """

    return float(euclidean(
        u=position_matrix[atom1_id],
        v=position_matrix[atom2_id]
    ))
